const fs = require('fs');
const path = require('path');
const webpack = require('webpack');

const HtmlWebpackPlugin = require('html-webpack-plugin');
const CopyWebpackPlugin = require('copy-webpack-plugin');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const MomentTimezoneDataPlugin = require('moment-timezone-data-webpack-plugin');
const BundleAnalyzerPlugin = require('webpack-bundle-analyzer').BundleAnalyzerPlugin;

const getClientEnvironment = require('./env');
const paths = require('./paths');

const publicPath = paths.servedPath;
const publicUrl = publicPath.slice(0, -1);
const env = getClientEnvironment(publicUrl);

const isJsonOut = process.argv.indexOf('--json') >= 0;
const isProdMode = env.raw.NODE_ENV === "production";
const isVendorModule = (module) => module.resource ? module.resource.includes('/node_modules/') : false;
const isIndexChunk = (chunks) => chunks.length === 1 && chunks[0].name === 'index';
const isMocksChunk = (chunks) => chunks.length === 1 && chunks[0].name === 'mocks';
const isIframeChunk = (chunks) => chunks.length === 1 && chunks[0].name === 'iframe';
const isCommonChunk = (chunks) => chunks.length > 1;

const outPath = paths.appBuild;
const sourcePath = paths.appSrc;
const staticPath = paths.appPublic;

const indexEntries = ['./index.tsx', './index.jsx', './index.js']
  .filter((it) => fs.existsSync(path.join(paths.appSrc, it)));
const indexMocksEntries = ['./index-mocks.tsx', './index-mocks.jsx', './index-mocks.js']
  .filter((it) => fs.existsSync(path.join(paths.appSrc, it)));
const iframeMocksEntries = ['./index-iframe.tsx', './index-iframe.jsx', './index-iframe.js']
  .filter((it) => fs.existsSync(path.join(paths.appSrc, it)));

module.exports = {
  context: sourcePath,
  entry: {
    index: [
      // '@babel/polyfill',
      'react-hot-loader/patch',
    ].concat(indexEntries),
    mocks: [
      // '@babel/polyfill',
      'react-hot-loader/patch',
    ].concat(indexMocksEntries),
    iframe: [
      // '@babel/polyfill',
      'react-hot-loader/patch',
    ].concat(iframeMocksEntries),
  },
  output: {
    path: outPath,
    publicPath: '',
    filename: isProdMode ? '[name].[chunkhash:8].js' : '[name].[hash:8].js',
  },
  target: 'web',
  resolve: {
    extensions: ['.js', '.jsx', '.ts', '.tsx'],
    mainFields: ['browser', 'main'],
    alias: {
      src: sourcePath,
      static: staticPath,
      ...getAliases(),
    },
  },
  module: {
    rules: [
      {
        test: /\.tsx?$/,
        use: [
          {
            loader: 'ts-loader',
            options: {
              silent: isJsonOut,
              transpileOnly: true,
              configFile: paths.appTsConfig,
            },
          },
        ],
      },
      {
        test: /\.tsx?$/,
        enforce: 'post',
        use: env.raw.REACT_APP_COVERAGE
          ? [{loader: 'istanbul-instrumenter-loader', options: {esModules: true}}]
          : [],
      },
      {
        test: /.+\.po$/,
        include: path.join(staticPath, 'translations'),
        use: [
          {
            loader: 'json-loader',
          },
          {
            loader: 'po-loader',
          },
        ],
      },
      {
        include: path.join(staticPath, 'mockfiles'),
        use: 'base64-loader',
      },
      {
        test: /\.less$/,
        use: [
          {
            loader: MiniCssExtractPlugin.loader,
            options: {
              hmr: !isProdMode,
              reloadAll: true,
            },
          },
          {
            loader: 'css-loader',
            options: {
              modules: false,
              sourceMap: true,
            },
          },
          {
            loader: 'less-loader',
            options: {
              paths: [paths.appPath],
              sourceMap: true,
            },
          },
        ],
      },
      {
        test: /\.css$/,
        use: [
          {
            loader: MiniCssExtractPlugin.loader,
            options: {
              hmr: !isProdMode,
              reloadAll: true,
            },
          },
          {
            loader: 'css-loader',
            options: {
              modules: false,
              sourceMap: true,
            },
          },
        ],
      },
      {
        test: /\.html$/,
        use: 'html-loader',
        exclude: staticPath,
      },
      {
        test: /\.(png|gif)$/,
        loader: 'url-loader',
        options: {name: 'images/[name].[ext]', limit: 10000},
        exclude: staticPath,
      },
      {
        test: /\.jpg$/,
        loader: 'file-loader',
        options: {name: 'images/[name].[ext]'},
        exclude: staticPath,
      },
      {
        test: /\.(eot|svg|ttf|woff|woff2)$/,
        loader: 'file-loader',
        options: {name: 'fonts/[name].[ext]'},
        exclude: staticPath,
      },
    ],
  },
  optimization: {
    splitChunks: {
      name: false,
      cacheGroups: {
        'vendors-iframe': {
          chunks: 'all',
          priority: -10,
          name: 'vendors-iframe',
          enforce: true,
          reuseExistingChunk: false,
          test: (module, chunks) => isVendorModule(module) && isIframeChunk(chunks),
        },
        'iframe': {
          chunks: 'all',
          priority: -10,
          name: 'iframe',
          enforce: true,
          reuseExistingChunk: false,
          test: (module, chunks) => !isVendorModule(module) && isIframeChunk(chunks),
        },

        'vendors-mocks': {
          chunks: 'all',
          priority: -10,
          name: 'vendors-mocks',
          enforce: true,
          reuseExistingChunk: false,
          test: (module, chunks) => isVendorModule(module) && isMocksChunk(chunks),
        },
        'mocks': {
          chunks: 'all',
          priority: -10,
          name: 'mocks',
          enforce: true,
          reuseExistingChunk: false,
          test: (module, chunks) => !isVendorModule(module) && isMocksChunk(chunks),
        },

        'vendors-index': {
          chunks: 'all',
          priority: -10,
          name: 'vendors-index',
          enforce: true,
          reuseExistingChunk: false,
          test: (module, chunks) => isVendorModule(module) && isIndexChunk(chunks),
        },
        'index': {
          chunks: 'all',
          priority: -10,
          name: 'index',
          enforce: true,
          reuseExistingChunk: false,
          test: (module, chunks) => !isVendorModule(module) && isIndexChunk(chunks),
        },

        'vendors-common': {
          chunks: 'all',
          priority: -10,
          name: 'vendors-common',
          enforce: true,
          reuseExistingChunk: false,
          test: (module, chunks) => isVendorModule(module) && isCommonChunk(chunks),
        },
        'common': {
          chunks: 'all',
          priority: -10,
          name: 'common',
          enforce: true,
          reuseExistingChunk: false,
          test: (module, chunks) => !isVendorModule(module) && isCommonChunk(chunks),
        },
      },
    },
  },
  plugins: getWebpackPlugins(),
  devServer: {
    contentBase: sourcePath,
    hot: true,
    stats: {
      warnings: false,
    },
    allowedHosts: [
      'dockerhost',
    ],
  },
  node: {
    // workaround for webpack-dev-server issue
    // https://github.com/webpack/webpack-dev-server/issues/60#issuecomment-103411179
    fs: 'empty',
    net: 'empty',
  },
  devtool: isProdMode ? "source-map" : "cheap-module-source-map",
};

function getWebpackPlugins() {
  let libraries = {};
  if (env.raw.REACT_APP_LIBRARIES) {
    try {
      libraries = JSON.parse(env.raw.REACT_APP_LIBRARIES);
      if (!libraries) {
        libraries = {};
      }
    } catch (e) {
    }
  }

  const plugins = [
    new webpack.NamedModulesPlugin(),
    new webpack.ProvidePlugin(libraries),
    new webpack.DefinePlugin(env.stringified),
    new webpack.IgnorePlugin(/^\.\/locale$/, /moment$/),
    new MomentTimezoneDataPlugin({
      matchZones: 'Europe/Zurich',
    }),
    new webpack.optimize.AggressiveMergingPlugin(),
    new MiniCssExtractPlugin({
      filename: isProdMode
          ? 'styles.[chunkhash:8].css'
          : 'styles.css',
    }),
    new HtmlWebpackPlugin({
      template: 'index.html',
      filename: 'index.html',
      inject: true,
      minify: false,
      chunks: ['vendors-common', 'vendors-index', 'common', 'index'],
      chunksSortMode: 'manual',
    }),
    new HtmlWebpackPlugin({
      template: 'index.html',
      filename: 'mocks.html',
      inject: true,
      minify: false,
      chunks: ['vendors-common', 'vendors-index', 'vendors-mocks', 'common', 'mocks', 'index'],
      chunksSortMode: 'manual',
    }),
    new HtmlWebpackPlugin({
      template: 'index.html',
      filename: 'iframe.html',
      inject: true,
      minify: false,
      chunks: ['vendors-common', 'vendors-index', 'vendors-iframe', 'common', 'iframe', 'index'],
      chunksSortMode: 'manual',
    }),
    new CopyWebpackPlugin([
      {
        from: path.join(staticPath, 'favicon.ico'),
        to: 'favicon.ico',
      },
      {
        from: path.join(staticPath, 'manifest.json'),
        to: 'manifest.json',
      },
      {
        from: path.join(staticPath, 'images'),
        to: 'images',
      },
      {
        context: path.resolve(paths.appPath, 'submodules'),
        from: '**/*',
        to: 'modules',
      },
      ...getStaticFiles(),
    ]),
  ];
  if (process.env.REACT_APP_ANALYZE) {
    plugins.push(new BundleAnalyzerPlugin());
  }
  return plugins;
}

function getAliases() {
  try {
    const aliases = env.raw.REACT_APP_ALIASES ? JSON.parse(env.raw.REACT_APP_ALIASES) : {};
    return Object
        .keys(aliases)
        .reduce((result, key) => ({
          ...result,
          [key]: path.resolve(paths.appPath, aliases[key])
        }), {});
  } catch (e) {
    return {};
  }
}

function getStaticFiles() {
  try {
    const files = env.raw.REACT_APP_STATIC_FILES ? JSON.parse(env.raw.REACT_APP_STATIC_FILES) : [];
    return files.map(({ from, to }) => ({ context: paths.appPath, from, to }));
  } catch (e) {
    return [];
  }
}
